import copy
from typing import Optional

from ..dialog.serialized_message import SerializedMessage
from ..dialog.names import SOURCES
from ..adapters.base import BaseAdapter, Context, Response


class VkAdapter(BaseAdapter):
    SOURCE = SOURCES.VK

    def __init__(
            self,
            suggest_cols='auto', suggest_screen=32, suggest_margin=1, suggest_max_len=40,
            suggest_max_cols=5, suggest_max_rows=10,
            suggest_add_payload=False,
            **kwargs
    ):
        super(VkAdapter, self).__init__(**kwargs)
        self.suggest_cols = suggest_cols
        self.suggest_screen = suggest_screen
        self.suggest_margin = suggest_margin
        self.suggest_max_len = suggest_max_len
        self.suggest_max_cols = suggest_max_cols
        self.suggest_max_rows = suggest_max_rows
        self.suggest_add_payload = suggest_add_payload

    def make_context(self, message, **kwargs) -> Context:
        uid = self.SOURCE + '__' + str(message.user_id)
        ctx = Context(
            user_object=None,
            message_text=message.text,
            metadata={},
            user_id=uid,
            session_id=uid,
            source=self.SOURCE,
            raw_message=message,
        )
        return ctx

    def make_response(self, response: Response, original_message=None, **kwargs):
        # todo: instead of a dict, use a class object as a response
        # todo: add multimedia, etc.
        result = {
            'text': response.text,
        }
        if response.suggests or response.links:
            buttons = []
            for i, link in enumerate(response.links):
                buttons.append({'action': {'type': 'open_link', 'label': link['title'], 'link': link['url']}})
            for i, suggest in enumerate(response.suggests):
                a = {'type': 'text', 'label': suggest}
                if self.suggest_add_payload:
                    a['payload'] = suggest
                buttons.append({'action': a})

            rows = []
            row_width = 0
            for i, button in enumerate(buttons):
                if self.suggest_cols == 'auto':
                    extra_width = len(button['action']['label']) + self.suggest_margin * 2
                    if len(rows) == 0 or row_width > 0 and row_width + extra_width > self.suggest_screen \
                            or len(rows[-1]) >= self.suggest_max_cols:
                        rows.append([])
                        row_width = extra_width
                    else:
                        row_width += extra_width
                else:
                    if i % self.suggest_cols == 0:
                        rows.append([])
                rows[-1].append(button)

            for row in rows:
                for button in row:
                    label = button['action']['label']
                    if len(label) > self.suggest_max_len:
                        button['action']['label'] = label[:(self.suggest_max_len - 1)] + '…'

            if self.suggest_max_rows:
                rows = self.squeeze_keyboard(rows)
                rows = rows[:self.suggest_max_rows]

            result['keyboard'] = {
                'one_time': True,
                'buttons': rows,
            }
        return result

    def squeeze_keyboard(self, rows):
        """ Shorten some buttons so that all buttons fit into the screen """
        for _ in range(100):
            if len(rows) <= self.suggest_max_rows:
                break
            # estimate free space in each row
            potentials = [
                sum([len(button['action']['label']) + self.suggest_margin * 2 for button in row])
                for row in rows
            ]
            # find two neighbor rows with most total free space
            best_pot = 1000
            best_i = 0
            for i in range(1, len(rows)):
                pot = potentials[i - 1] + potentials[i]
                if pot <= best_pot and len(rows[i-1]) + len(rows[i]) <= self.suggest_max_cols:
                    best_pot, best_i = pot, i
            if best_i == 0:
                break
            # calculate button sizes if the rows are merged
            to_reduce = best_pot - self.suggest_screen
            new_buttons = [copy.deepcopy(b) for row in rows[best_i - 1:best_i + 1] for b in row]
            sizes = [len(b['action']['label']) for b in new_buttons]
            while to_reduce > 0 and max(sizes) > 3:
                new_sizes = []
                maxsize = max(sizes)
                for s in sizes:
                    if s == maxsize and to_reduce:
                        s -= 1
                        to_reduce -= 1
                    new_sizes.append(s)
                sizes = new_sizes
            # actually, reduce the buttons
            for new_size, b in zip(sizes, new_buttons):
                n = len(b['action']['label'])
                if n > new_size:
                    b['action']['label'] = b['action']['label'][:new_size - 1] + '…'
            # update the rows
            rows = rows[:best_i - 1] + [new_buttons] + rows[best_i + 1:]
        return rows

    def serialize_context(self, context: Context, data=None, **kwargs) -> Optional[SerializedMessage]:
        serializable_message = {'message': context.raw_message and context.raw_message.to_json()}
        return super(VkAdapter, self).serialize_context(context=context, data=serializable_message, **kwargs)



from dialogic.nlg import reply_markup, controls
from dialogic.dialog.names import COMMANDS


class Response:
    def __init__(
            self,
            text,
            suggests=None,
            commands=None,
            voice=None,
            links=None,
            image_id=None,
            image_url=None,
            sound_url=None,
            gallery=None,
            image=None,
            user_object=None,
            raw_response=None,
            confidence=0.5,
            label=None,
            handler=None,
            rich_text=None,
            show_item_meta=None,
            no_response=False,  # whether the response (in messenger) should be emtpy
            attachment_filename=None,
            extra_directives=None,
            should_listen=None,
    ):
        self.text = text
        self.suggests = suggests or []
        self.commands = commands or []
        self.voice = voice if voice is not None else text
        self.links = links or []
        self.updated_user_object = user_object
        self.confidence = confidence
        self.image_id = image_id
        self.image_url = image_url  # todo: support them in Facebook as well
        self.sound_url = sound_url
        self.gallery = gallery
        assert self.gallery is None or isinstance(self.gallery, (controls.Gallery, controls.ImageGallery))
        self.image = image
        assert self.image is None or isinstance(self.image, controls.BigImage)
        self.raw_response = raw_response
        self.handler = handler
        self.label = label
        if rich_text:
            self.set_text(rich_text)
        self.show_item_meta = show_item_meta
        self.no_response = no_response
        self.attachment_filename = attachment_filename
        self.extra_directives = extra_directives  # a dictionary of directives for Alice
        self.should_listen = should_listen

    @property
    def user_object(self):
        # make it readonly, for clarity
        return self.updated_user_object

    def set_rich_text(self, rich_text):
        parser = reply_markup.TTSParser()
        try:
            parser.feed(rich_text)
        except ValueError as e:
            raise ValueError('Got error "{}" while parsing text "{}"'.format(e, rich_text))
        parser.close()
        self.text = parser.get_text()
        self.voice = parser.get_voice()
        self.links.extend(parser.get_links())
        if parser.get_image_id():
            self.image_id = parser.get_image_id()
        if parser.get_image_url():
            self.image_url = parser.get_image_url()
        return self

    def set_text(self, text_and_voice):
        # this method name is deprecated
        return self.set_rich_text(rich_text=text_and_voice)

    def add_link(self, title, url, hide=False):
        self.links.append({
            'title': title,
            'url': url,
            'hide': hide,
        })

    @property
    def has_exit_command(self) -> bool:
        return bool(self.commands and COMMANDS.EXIT in self.commands)



from .cascade import DialogTurn


def is_morning_show(turn: DialogTurn) -> bool:
    if not turn.ctx.yandex or not turn.ctx.yandex.request:
        return False
    r = turn.ctx.yandex.request
    if r.type != 'Show.Pull':
        return False
    return r.show_type == 'MORNING'
